import unittest
from typing import Any

from api.routes.web import (
    ExportProjectParticipantFields,
    ProjectParticipantGridFilter,
    prepare_participants_for_export,
)
from db.python.filters import GenericFilter
from db.python.layers import (
    AssayLayer,
    ParticipantLayer,
    ProjectInsightsLayer,
    SampleLayer,
    SequencingGroupLayer,
    WebLayer,
)
from db.python.tables.participant import ParticipantFilter
from models.enums.web import MetaSearchEntityPrefix
from models.models import (
    PRIMARY_EXTERNAL_ORG,
    Assay,
    AssayInternal,
    AssayUpsertInternal,
    FamilySimple,
    NestedParticipant,
    NestedSample,
    ParticipantUpsertInternal,
    ProjectSummaryInternal,
    SampleUpsertInternal,
    SequencingGroupUpsertInternal,
    WebProject,
)
from models.models.sequencing_group import NestedSequencingGroup
from models.models.web import (
    ProjectParticipantGridField,
    ProjectParticipantGridFilterType,
    ProjectParticipantGridResponse,
)
from models.utils.sample_id_format import sample_id_format, sample_id_transform_to_raw
from models.utils.sequencing_group_id_format import (
    sequencing_group_id_format,
    sequencing_group_id_transform_to_raw,
)
from test.testbase import TEST_PROJECT_NAME, DbIsolatedTest, run_as_sync

default_assay_meta = {
    'sequencing_type': 'genome',
    'sequencing_technology': 'short-read',
    'sequencing_platform': 'illumina',
}

DEFAULT_FAMILY_FIELDS = [
    ProjectParticipantGridField(
        key='external_ids',
        label='Family ID',
        is_visible=True,
        filter_key='external_id',
    )
]

DEFAULT_PARTICIPANT_FIELDS = [
    ProjectParticipantGridField(
        key='id',
        label='Participant ID',
        is_visible=True,
        filter_key='id',
        filter_types=[
            ProjectParticipantGridFilterType.eq,
            ProjectParticipantGridFilterType.neq,
            ProjectParticipantGridFilterType.gt,
            ProjectParticipantGridFilterType.gte,
            ProjectParticipantGridFilterType.lt,
            ProjectParticipantGridFilterType.lte,
        ],
    ),
    ProjectParticipantGridField(
        key='external_ids',
        label='External Participant ID',
        is_visible=True,
        filter_key='external_id',
    ),
    ProjectParticipantGridField(
        key='reported_sex',
        label='Reported sex',
        is_visible=False,
        filter_key='reported_sex',
    ),
    ProjectParticipantGridField(
        key='reported_gender',
        label='Reported gender',
        is_visible=False,
        filter_key='reported_gender',
        filter_types=None,
    ),
    ProjectParticipantGridField(
        key='karyotype',
        label='Karyotype',
        is_visible=False,
        filter_key='karyotype',
        filter_types=None,
    ),
]

DEFAULT_SAMPLE_FIELDS = [
    ProjectParticipantGridField(
        key='id',
        label='Sample ID',
        is_visible=True,
        filter_key='id',
        filter_types=[
            ProjectParticipantGridFilterType.eq,
            ProjectParticipantGridFilterType.neq,
        ],
    ),
    ProjectParticipantGridField(
        key='external_ids',
        label='External Sample ID',
        is_visible=True,
        filter_key='external_id',
    ),
    ProjectParticipantGridField(
        key='sample_root_id',
        label='Root Sample ID',
        is_visible=False,
        filter_key='sample_root_id',
    ),
    ProjectParticipantGridField(
        key='sample_parent_id',
        label='Parent Sample ID',
        is_visible=False,
        filter_key='sample_root_id',
    ),
    ProjectParticipantGridField(
        key='type', label='Type', is_visible=True, filter_key=None, filter_types=None
    ),
    ProjectParticipantGridField(
        key='created_date',
        label='Created date',
        is_visible=True,
        filter_key=None,
        filter_types=None,
    ),
]

DEFAULT_SEQ_GROUP_FIELDS = [
    ProjectParticipantGridField(
        key='id',
        label='Sequencing Group ID',
        is_visible=True,
        filter_key='id',
        filter_types=[
            ProjectParticipantGridFilterType.eq,
            ProjectParticipantGridFilterType.neq,
        ],
    ),
    ProjectParticipantGridField(
        key='type',
        label='Type',
        is_visible=True,
        filter_key='type',
    ),
    ProjectParticipantGridField(
        key='technology',
        label='Technology',
        is_visible=True,
        filter_key='technology',
    ),
    ProjectParticipantGridField(
        key='platform',
        label='Platform',
        is_visible=True,
        filter_key='platform',
    ),
]

DEFAULT_ASSAY_TYPES = [
    ProjectParticipantGridField(
        key='id',
        label='Assay ID',
        is_visible=True,
        filter_key='id',
        filter_types=[
            ProjectParticipantGridFilterType.eq,
            ProjectParticipantGridFilterType.neq,
        ],
    ),
    ProjectParticipantGridField(
        key='type', label='Type', is_visible=True, filter_key='type', filter_types=None
    ),
    ProjectParticipantGridField(
        key='meta.sequencing_type',
        label='sequencing_type',
        is_visible=False,
        filter_key='meta.sequencing_type',
        filter_types=None,
    ),
    ProjectParticipantGridField(
        key='meta.sequencing_platform',
        label='sequencing_platform',
        is_visible=False,
        filter_key='meta.sequencing_platform',
        filter_types=None,
    ),
    ProjectParticipantGridField(
        key='meta.sequencing_technology',
        label='sequencing_technology',
        is_visible=False,
        filter_key='meta.sequencing_technology',
        filter_types=None,
    ),
    ProjectParticipantGridField(
        key='meta.batch',
        label='batch',
        is_visible=True,
        filter_key='meta.batch',
        filter_types=None,
    ),
]


def data_to_class(data: dict | list) -> dict | list:
    """Convert the data into it's class using the _class field"""
    if isinstance(data, list):
        return [data_to_class(x) for x in data]

    if not isinstance(data, dict):
        return data

    cls = data.pop('_class', None)
    mapped_data = {k: data_to_class(v) for k, v in data.items()}

    if isinstance(cls, type):
        return cls(**mapped_data)

    return mapped_data


def merge(d1: dict, d2: dict):
    """Merges two dictionaries"""
    return dict(d1, **d2)


def get_test_participant():
    """Do it like this to avoid an upsert writing the test value"""
    return ParticipantUpsertInternal(
        external_ids={PRIMARY_EXTERNAL_ORG: 'Demeter'},
        meta={},
        samples=[
            SampleUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'sample_id001'},
                meta={},
                type='blood',
                sequencing_groups=[
                    SequencingGroupUpsertInternal(
                        type='genome',
                        technology='short-read',
                        platform='illumina',
                        assays=[
                            AssayUpsertInternal(
                                type='sequencing',
                                meta={
                                    'reads': [
                                        {
                                            'basename': 'sample_id001.filename-R1.fastq.gz',
                                            'checksum': None,
                                            'class': 'File',
                                            'location': '/path/to/sample_id001.filename-R1.fastq.gz',
                                            'size': 111,
                                        },
                                        {
                                            'basename': 'sample_id001.filename-R2.fastq.gz',
                                            'checksum': None,
                                            'class': 'File',
                                            'location': '/path/to/sample_id001.filename-R2.fastq.gz',
                                            'size': 111,
                                        },
                                    ],
                                    'reads_type': 'fastq',
                                    'batch': 'M001',
                                    **default_assay_meta,
                                },
                            ),
                        ],
                    )
                ],
            )
        ],
    )


def get_test_participant_2():
    """Do it like this to avoid an upsert writing the test value"""
    return ParticipantUpsertInternal(
        external_ids={PRIMARY_EXTERNAL_ORG: 'Meter'},
        meta={},
        samples=[
            SampleUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'sample_id002'},
                meta={},
                type='blood',
                sequencing_groups=[
                    SequencingGroupUpsertInternal(
                        type='genome',
                        technology='short-read',
                        platform='Illumina',
                        assays=[
                            AssayUpsertInternal(
                                type='sequencing',
                                meta={
                                    'reads': [
                                        {
                                            'basename': 'sample_id002.filename-R1.fastq.gz',
                                            'checksum': None,
                                            'class': 'File',
                                            'location': '/path/to/sample_id002.filename-R1.fastq.gz',
                                            'size': 112,
                                        },
                                        {
                                            'basename': 'sample_id002.filename-R2.fastq.gz',
                                            'checksum': None,
                                            'class': 'File',
                                            'location': '/path/to/sample_id002.filename-R2.fastq.gz',
                                            'size': 112,
                                        },
                                    ],
                                    'reads_type': 'fastq',
                                    'batch': 'M001',
                                    'field with spaces': 'field with spaces',
                                    **default_assay_meta,
                                },
                            ),
                        ],
                    ),
                ],
            )
        ],
    )


SINGLE_PARTICIPANT_SUMMARY_RESULT = ProjectSummaryInternal(
    project=WebProject(id=1, name=TEST_PROJECT_NAME, meta={}, dataset='test'),
    total_samples=1,
    total_participants=1,
    total_sequencing_groups=1,
    total_assays=1,
    cram_seqr_stats={
        'genome': {
            'Sequences': '1',
            'Crams': '0',
            'Seqr': '0',
        }
    },
    batch_sequencing_group_stats={'M001': {'genome': '1'}},
    seqr_links={},
    seqr_sync_types=[],
)

SINGLE_PARTICIPANT_QUERY_RESULT = ProjectParticipantGridResponse(
    participants=[],
    total_results=1,
    fields={
        MetaSearchEntityPrefix.FAMILY: [
            ProjectParticipantGridField(
                key='external_ids', label='Family ID', is_visible=True
            )
        ],
        MetaSearchEntityPrefix.PARTICIPANT: [
            ProjectParticipantGridField(
                key='external_ids', label='Participant ID', is_visible=True
            ),
        ],
        MetaSearchEntityPrefix.SAMPLE: [
            ProjectParticipantGridField(key='meta.skey', label='', is_visible=True),
            ProjectParticipantGridField(key='external_ids', label='', is_visible=True),
        ],
        MetaSearchEntityPrefix.SEQUENCING_GROUP: [
            ProjectParticipantGridField(key='type', label='', is_visible=True),
            ProjectParticipantGridField(key='meta.sgkey', label='', is_visible=True),
        ],
        MetaSearchEntityPrefix.ASSAY: [
            ProjectParticipantGridField(key='type', label='', is_visible=True),
            ProjectParticipantGridField(key='meta.akey', label='', is_visible=True),
        ],
    },
)


class TestWeb(DbIsolatedTest):
    """Test web class containing web endpoints"""

    maxDiff = None

    @run_as_sync
    async def setUp(self) -> None:
        super().setUp()
        self.webl = WebLayer(self.connection)
        self.partl = ParticipantLayer(self.connection)
        self.pil = ProjectInsightsLayer(self.connection)
        self.sampl = SampleLayer(self.connection)
        self.seql = AssayLayer(self.connection)

    @run_as_sync
    async def test_project_summary_empty(self):
        """Test getting the summary for a project"""
        result = await self.webl.get_project_summary()

        # Expect an empty project
        expected = ProjectSummaryInternal(
            project=WebProject(id=1, name=self.project_name, meta={}, dataset='test'),
            total_samples=0,
            total_participants=0,
            total_sequencing_groups=0,
            total_assays=0,
            batch_sequencing_group_stats={},
            cram_seqr_stats={},
            seqr_links={},
            seqr_sync_types=[],
        )

        self.assertEqual(expected, result)

    @run_as_sync
    async def test_project_summary_single_entry(self):
        """Test project summary with a single participant with all fields"""
        # Now add a participant with a sample and sequence
        await self.partl.upsert_participants(participants=[get_test_participant()])

        result = await self.webl.get_project_summary()
        self.assertDataclassEqual(SINGLE_PARTICIPANT_SUMMARY_RESULT, result)

    @run_as_sync
    async def test_project_summary_to_external(self):
        """Test project summary to_external function"""
        # Now add a participant with a sample and sequence
        await self.partl.upsert_participants(participants=[get_test_participant()])

        summary = await self.webl.get_project_summary()
        self.assertEqual(
            SINGLE_PARTICIPANT_SUMMARY_RESULT.to_external(),
            summary.to_external(),
        )
        internal_participants = await self.webl.query_participants(
            ParticipantFilter(), limit=None
        )

        ex_result = ProjectParticipantGridResponse.from_params(
            participants=internal_participants,
            total_results=1,
            filter_fields=ProjectParticipantGridFilter(),
        )

        assert isinstance(internal_participants[0].samples, list)
        self.assertIsInstance(internal_participants[0].samples[0].id, int)
        self.assertIsInstance(ex_result.participants[0].samples[0].id, str)
        self.assertEqual(
            sample_id_transform_to_raw(ex_result.participants[0].samples[0].id),
            internal_participants[0].samples[0].id,
        )

        assert isinstance(internal_participants[0].samples[0].sequencing_groups, list)
        assert isinstance(ex_result.participants[0].samples[0].sequencing_groups, list)

        self.assertIsInstance(
            internal_participants[0].samples[0].sequencing_groups[0].id, int
        )
        self.assertIsInstance(
            ex_result.participants[0].samples[0].sequencing_groups[0].id, str
        )
        self.assertEqual(
            sequencing_group_id_transform_to_raw(
                ex_result.participants[0].samples[0].sequencing_groups[0].id
            ),
            internal_participants[0].samples[0].sequencing_groups[0].id,
        )

        assert isinstance(
            internal_participants[0].samples[0].sequencing_groups[0].assays, list
        )
        assert isinstance(
            ex_result.participants[0].samples[0].sequencing_groups[0].assays, list
        )
        self.assertIsInstance(
            internal_participants[0].samples[0].sequencing_groups[0].assays[0],
            AssayInternal,
        )
        self.assertIsInstance(
            ex_result.participants[0].samples[0].sequencing_groups[0].assays[0], Assay
        )

    @run_as_sync
    async def project_summary_with_filter_with_results(self):
        """Project grid but with test filter, that shows results"""
        await self.partl.upsert_participants(participants=[get_test_participant()])

        pfilter = ProjectParticipantGridFilter(
            assay=ProjectParticipantGridFilter.ParticipantGridAssayFilter(
                meta={'batch': GenericFilter[Any](startswith='M001')}
            )
        )

        nested_participants = await self.webl.query_participants(
            pfilter.to_internal(project=self.project_id), limit=None
        )
        result = ProjectParticipantGridResponse.from_params(
            participants=nested_participants,
            total_results=1,
            filter_fields=pfilter,
        )
        self.assertEqual(1, len(nested_participants))
        result.participants = []
        self.assertEqual(SINGLE_PARTICIPANT_QUERY_RESULT, result)

    @run_as_sync
    async def project_summary_with_filter_no_results(self):
        """Project grid but with test filter, that doesn't have results"""
        await self.partl.upsert_participants(participants=[get_test_participant()])
        pfilter = ProjectParticipantGridFilter(
            assay=ProjectParticipantGridFilter.ParticipantGridAssayFilter(
                meta={'batch': GenericFilter[Any](startswith='M002')}
            )
        )

        nested_participants = await self.webl.query_participants(
            pfilter.to_internal(project=self.project_id), limit=None
        )
        self.assertEqual(0, len(nested_participants))

        result = ProjectParticipantGridResponse.from_params(
            participants=nested_participants,
            total_results=0,
            filter_fields=pfilter,
        )
        result.participants = []
        self.assertEqual(SINGLE_PARTICIPANT_QUERY_RESULT, result)

        empty_result = ProjectParticipantGridResponse(
            total_results=0, participants=[], fields={}
        )

        self.assertEqual(empty_result, result)

    @run_as_sync
    async def test_project_summary_multiple_participants(self):
        """Try with multiple participants as some extra security"""
        await self.partl.upsert_participants(
            participants=[get_test_participant(), get_test_participant_2()]
        )

        expected_summary = ProjectSummaryInternal(
            project=WebProject(id=1, name=self.project_name, meta={}, dataset='test'),
            total_samples=2,
            total_participants=2,
            total_sequencing_groups=2,
            total_assays=2,
            cram_seqr_stats={
                'genome': {
                    'Sequences': '2',
                    'Crams': '0',
                    'Seqr': '0',
                }
            },
            batch_sequencing_group_stats={'M001': {'genome': '2'}},
            seqr_links={},
            seqr_sync_types=[],
        )

        expected_fields = {
            MetaSearchEntityPrefix.FAMILY: DEFAULT_FAMILY_FIELDS,
            MetaSearchEntityPrefix.PARTICIPANT: DEFAULT_PARTICIPANT_FIELDS,
            MetaSearchEntityPrefix.SAMPLE: DEFAULT_SAMPLE_FIELDS,
            MetaSearchEntityPrefix.SEQUENCING_GROUP: DEFAULT_SEQ_GROUP_FIELDS,
            MetaSearchEntityPrefix.ASSAY: [
                *DEFAULT_ASSAY_TYPES,
                ProjectParticipantGridField(
                    key='meta.reads_type',
                    label='reads_type',
                    is_visible=True,
                    filter_key='meta.reads_type',
                    filter_types=None,
                ),
                ProjectParticipantGridField(
                    key='meta.reads',
                    label='reads',
                    is_visible=False,
                    filter_key='meta.reads',
                    filter_types=None,
                ),
                ProjectParticipantGridField(
                    key='meta.field with spaces',
                    label='field with spaces',
                    is_visible=True,
                    filter_key='meta.field with spaces',
                    filter_types=None,
                ),
            ],
        }

        summary = await self.webl.get_project_summary()

        self.assertEqual(expected_summary, summary)

        nested_participants = await self.webl.query_participants(
            ParticipantFilter(), limit=None
        )
        self.assertEqual(2, len(nested_participants))
        result = ProjectParticipantGridResponse.from_params(
            participants=nested_participants,
            total_results=2,
            filter_fields=ProjectParticipantGridFilter(),
        )
        # make it easier to test
        result.participants = []

        self.maxDiff = None
        # comparison
        for k, expected_field in expected_fields.items():
            sorted_expected_fields = sorted(
                expected_field, key=lambda x: x.key
            )  # sort by key
            sorted_result_fields = sorted(
                result.fields[k], key=lambda x: x.key
            )  # sort by key
            self.assertEqual(
                sorted_expected_fields,
                sorted_result_fields,
                msg=f'Fields for category {k} did not match',
            )

    @run_as_sync
    async def test_project_summary_multiple_participants_and_filter(self):
        """Try with multiple participants as some extra security"""
        await self.partl.upsert_participants(
            participants=[get_test_participant(), get_test_participant_2()]
        )

        expected_fields = {
            MetaSearchEntityPrefix.FAMILY: DEFAULT_FAMILY_FIELDS,
            MetaSearchEntityPrefix.PARTICIPANT: DEFAULT_PARTICIPANT_FIELDS,
            MetaSearchEntityPrefix.SAMPLE: DEFAULT_SAMPLE_FIELDS,
            MetaSearchEntityPrefix.SEQUENCING_GROUP: DEFAULT_SEQ_GROUP_FIELDS,
            MetaSearchEntityPrefix.ASSAY: [
                *DEFAULT_ASSAY_TYPES,
                ProjectParticipantGridField(
                    key='meta.reads_type',
                    label='reads_type',
                    is_visible=True,
                    filter_key='meta.reads_type',
                    filter_types=None,
                ),
                ProjectParticipantGridField(
                    key='meta.reads',
                    label='reads',
                    is_visible=False,
                    filter_key='meta.reads',
                    filter_types=None,
                ),
                ProjectParticipantGridField(
                    key='meta.field with spaces',
                    label='field with spaces',
                    is_visible=True,
                    filter_key='meta.field with spaces',
                    filter_types=None,
                ),
            ],
        }

        pfilter = ProjectParticipantGridFilter(
            sample=ProjectParticipantGridFilter.ParticipantGridSampleFilter(
                external_id=GenericFilter[str](contains='_id002')
            )
        )

        nested_participants = await self.webl.query_participants(
            pfilter.to_internal(project=self.project_id), limit=None
        )

        self.assertEqual(1, len(nested_participants))
        result = ProjectParticipantGridResponse.from_params(
            nested_participants,
            filter_fields=pfilter,
            total_results=2,
        )
        result.participants = []

        for k, expected_field in expected_fields.items():
            sorted_expected_fields = sorted(
                expected_field, key=lambda x: x.key
            )  # sort by key
            sorted_result_fields = sorted(
                result.fields[k], key=lambda x: x.key
            )  # sort by key
            self.assertEqual(
                sorted_expected_fields,
                sorted_result_fields,
                msg=f'Fields for category {k} did not match',
            )

    @run_as_sync
    async def test_field_with_space(self):
        """Test filtering on a meta field with spaces"""
        await self.partl.upsert_participants(
            participants=[get_test_participant(), get_test_participant_2()]
        )

        pfilter = ProjectParticipantGridFilter(
            assay=ProjectParticipantGridFilter.ParticipantGridAssayFilter(
                meta={'field with spaces': GenericFilter[Any](contains='field wi')}
            )
        )
        nested_participants = await self.webl.query_participants(
            pfilter.to_internal(project=self.project_id), limit=None
        )

        self.assertEqual(1, len(nested_participants))

    @run_as_sync
    async def test_project_summary_inactive_sequencing_group(self):
        """
        Insert a sequencing-group, archive it, then check that the summary
        doesn't return that sequencing group
        """
        participants = await self.partl.upsert_participants(
            participants=[get_test_participant()]
        )
        assert (
            isinstance(participants, list)
            and isinstance(participants[0].samples, list)
            and isinstance(participants[0].samples[0].sequencing_groups, list)
        )
        sg = participants[0].samples[0].sequencing_groups[0]
        assert isinstance(sg.assays, list)
        assay_ids = [a.id for a in sg.assays if a.id]
        sglayer = SequencingGroupLayer(self.connection)
        assert sg.id
        new_sg_id = await sglayer.recreate_sequencing_group_with_new_assays(
            sequencing_group_id=sg.id,
            assays=assay_ids,
            meta={'new-meta': 'value'},
        )

        participants = await self.webl.query_participants(
            ParticipantFilter(), limit=None
        )
        summary_sgs = participants[0].samples[0].sequencing_groups
        self.assertEqual(1, len(summary_sgs))
        self.assertEqual(new_sg_id, summary_sgs[0].id)


class WebNonDBTests(unittest.TestCase):
    """Handy place to put tests that don't require a database"""

    def test_nested_participant_to_rows(self):
        """Test nested participant to flat rows with a projection"""
        participant = NestedParticipant(
            id=1,
            external_ids={PRIMARY_EXTERNAL_ORG: 'pex1'},
            meta={'pkey': 'value'},
            families=[FamilySimple(id=-2, external_ids={'family_org': 'fex1'})],
            samples=[
                NestedSample(
                    id='xpgA',
                    external_ids={PRIMARY_EXTERNAL_ORG: 'sex1', 'external_org': 'ex02'},
                    meta={'skey': 'svalue'},
                    type='blood',
                    created_date='2021-01-01',
                    non_sequencing_assays=[],
                    sample_root_id=None,
                    sample_parent_id=None,
                    sequencing_groups=[
                        NestedSequencingGroup(
                            id='cpgA',
                            type='genome',
                            external_ids={'sgex1': 'sgex1'},
                            technology='short-read',
                            platform='illumina',
                            meta={'sgkey': 'sgvalue'},
                            assays=[
                                Assay(
                                    id=-1,
                                    type='sequencing',
                                    external_ids={'ex1': 'ex1'},
                                    sample_id='xpgA',
                                    meta={'akey': 'avalue'},
                                ),
                                Assay(
                                    id=-2,
                                    type='sequencing',
                                    external_ids={'ex1': 'ex2'},
                                    sample_id='xpgA',
                                    meta={'akey': 'avalue2'},
                                ),
                            ],
                        )
                    ],
                )
            ],
        )
        fields = ExportProjectParticipantFields(
            fields={
                MetaSearchEntityPrefix.FAMILY: [
                    ProjectParticipantGridField(
                        key='external_ids', label='', is_visible=True
                    )
                ],
                MetaSearchEntityPrefix.PARTICIPANT: [
                    ProjectParticipantGridField(
                        key='external_ids', label='', is_visible=True
                    ),
                    ProjectParticipantGridField(
                        key='meta.pkey', label='', is_visible=True
                    ),
                ],
                MetaSearchEntityPrefix.SAMPLE: [
                    ProjectParticipantGridField(
                        key='meta.skey', label='', is_visible=True
                    ),
                    ProjectParticipantGridField(
                        key='external_ids', label='', is_visible=True
                    ),
                ],
                MetaSearchEntityPrefix.SEQUENCING_GROUP: [
                    ProjectParticipantGridField(key='type', label='', is_visible=True),
                    ProjectParticipantGridField(
                        key='meta.sgkey', label='', is_visible=True
                    ),
                ],
                MetaSearchEntityPrefix.ASSAY: [
                    ProjectParticipantGridField(key='type', label='', is_visible=True),
                    ProjectParticipantGridField(
                        key='meta.akey', label='', is_visible=True
                    ),
                ],
            }
        )

        i = prepare_participants_for_export([participant], fields)
        headers = next(i)
        rows = list(i)
        self.assertTupleEqual(
            headers,
            (
                'family.external_ids',
                'participant.external_ids',
                'participant.meta.pkey',
                'sample.meta.skey',
                'sample.external_ids',
                'sequencing_group.type',
                'sequencing_group.meta.sgkey',
                'assay.type',
                'assay.meta.akey',
            ),
        )
        non_sg_keys = (
            'family_org: fex1',
            'pex1',
            'value',
            'svalue',
            'sex1, external_org: ex02',
            'genome',
            'sgvalue',
        )
        expected = [
            (*non_sg_keys, 'sequencing', 'avalue'),
            (*non_sg_keys, 'sequencing', 'avalue2'),
        ]

        self.assertEqual(2, len(rows))

        self.assertListEqual(expected, rows)

    def test_project_participant_grid_filter(self):
        """
        Test every filter, and make sure it's being converted
        to the internal model correctly
        """
        f_id = 1
        p_id = 2
        s_id = 3
        s_id_ext = sample_id_format(s_id)
        sg_id = 4
        sg_id_ext = sequencing_group_id_format(sg_id)
        big_filter = ProjectParticipantGridFilter(
            family=ProjectParticipantGridFilter.ParticipantGridFamilyFilter(
                id=GenericFilter[int](contains=f_id)
            ),
            participant=ProjectParticipantGridFilter.ParticipantGridParticipantFilter(
                id=GenericFilter[int](contains=p_id),
                meta={'pmeta': GenericFilter[Any](contains='pm')},
                external_id=GenericFilter[str](contains='e'),
            ),
            sample=ProjectParticipantGridFilter.ParticipantGridSampleFilter(
                id=GenericFilter[str](contains=s_id_ext),
                type=GenericFilter[str](contains='t'),
                external_id=GenericFilter[str](contains='e'),
                meta={'smeta': GenericFilter[Any](contains='sm')},
            ),
            sequencing_group=ProjectParticipantGridFilter.ParticipantGridSequencingGroupFilter(
                id=GenericFilter[str](contains=sg_id_ext),
                type=GenericFilter[str](contains='t'),
                external_id=GenericFilter[str](contains='e'),
                meta={'sgmeta': GenericFilter[Any](contains='sg')},
                technology=GenericFilter[str](contains='t'),
                platform=GenericFilter[str](contains='p'),
            ),
            assay=ProjectParticipantGridFilter.ParticipantGridAssayFilter(
                id=GenericFilter[int](contains=5),
                type=GenericFilter[str](contains='t'),
                external_id=GenericFilter[str](contains='e'),
                meta={'ameta': GenericFilter[Any](contains='a')},
            ),
        )

        internal_filter = big_filter.to_internal(project=1)

        # participant internal
        self.assertEqual(internal_filter.id.contains, p_id)
        self.assertEqual(internal_filter.meta['pmeta'].contains, 'pm')
        self.assertEqual(internal_filter.external_id.contains, 'e')

        # family internal
        self.assertEqual(internal_filter.family.id.contains, f_id)

        # sample internal
        self.assertEqual(internal_filter.sample.id.contains, s_id)
        self.assertEqual(internal_filter.sample.type.contains, 't')
        self.assertEqual(internal_filter.sample.external_id.contains, 'e')
        self.assertEqual(internal_filter.sample.meta['smeta'].contains, 'sm')

        # sequencing group internal
        self.assertEqual(internal_filter.sequencing_group.id.contains, sg_id)
        self.assertEqual(internal_filter.sequencing_group.type.contains, 't')
        self.assertEqual(internal_filter.sequencing_group.external_id.contains, 'e')
        self.assertEqual(internal_filter.sequencing_group.meta['sgmeta'].contains, 'sg')
        self.assertEqual(internal_filter.sequencing_group.technology.contains, 't')
        self.assertEqual(internal_filter.sequencing_group.platform.contains, 'p')

        # assay internal
        self.assertEqual(internal_filter.assay.id.contains, 5)
        self.assertEqual(internal_filter.assay.type.contains, 't')
        self.assertEqual(internal_filter.assay.external_id.contains, 'e')
        self.assertEqual(internal_filter.assay.meta['ameta'].contains, 'a')



import unittest

from metamist.models import AssayUpsert


class TestMetamistGeneratedLibrary(unittest.TestCase):
    """
    Test the generated library, mostly instantiating models to check we don't regress.
    """

    def test_create_assay_with_id(self):
        """Just test instantiating the model"""
        _ = AssayUpsert(id=1, external_ids={'blah': 'blah'})

    def test_create_assay_no_id(self):
        """Test instantiating the model without an id"""
        _ = AssayUpsert(id=None, external_ids={'blah': 'blah'})



from tempfile import TemporaryFile

from fastapi import HTTPException, UploadFile

from api.routes import family
from test.testbase import DbIsolatedTest, run_as_sync


class TestFamilyImportEndpoint(DbIsolatedTest):
    """Family testing methods"""

    @run_as_sync
    async def test_import_families_empty_file(self):
        """
        Test importing families from a file where the file is empty.
        """
        with TemporaryFile(prefix='test', suffix='.tsv') as f:
            emptyTestFile = UploadFile(f)

            # Test has_header = true
            with self.assertRaises(HTTPException) as context:
                _ = await family.import_families(
                    emptyTestFile,
                    has_header=True,
                    delimiter='\t',
                    connection=self.connection,
                )
            self.assertEqual(context.exception.status_code, 400)
            self.assertIn(
                'A header was expected but file is empty.', context.exception.detail
            )

            # Test no op when has_header = false
            response = await family.import_families(
                emptyTestFile,
                has_header=False,
                delimiter='\t',
                connection=self.connection,
            )
            self.assertDictEqual(
                response,
                {
                    'success': True,
                    'warnings': ['Submitted file was empty'],
                },
            )

    @run_as_sync
    async def test_import_families_header_no_content(self):
        """
        Test importing families from a file with a header but no data.
        """
        with TemporaryFile(mode='wb+', prefix='test', suffix='.tsv') as f:
            f.write(b'Some\ttest\theader\twithout\tdata\n')
            f.seek(0)
            headerOnlyFile = UploadFile(f)

            # Test has_header = true
            response = await family.import_families(
                headerOnlyFile,
                has_header=True,
                delimiter='\t',
                connection=self.connection,
            )
            self.assertDictEqual(
                response,
                {
                    'success': True,
                    'warnings': ['Submitted file contained a header with no data'],
                },
            )

    @run_as_sync
    async def test_import_families_valid_data(self):
        """
        Test importing families with valid file contents.
        """
        data = [
            ['familyid', 'description', 'phenotype'],
            ['Smith', 'Blacksmiths', 'burnt'],
            ['Jones', 'From Wales', 'sings well'],
            ['Taylor', 'Post Norman', 'sews'],
        ]
        fileContent = '\n'.join(['\t'.join(row) for row in data]).encode(
            encoding='utf-8-sig'
        )

        with TemporaryFile(mode='wb+', prefix='test', suffix='.tsv') as f:
            f.write(fileContent)
            f.seek(0)
            testFile = UploadFile(f)

            # Test has_header = true
            response = await family.import_families(
                testFile,
                has_header=True,
                delimiter='\t',
                connection=self.connection,
            )
            self.assertDictEqual(response, {'success': True})

            f.seek(0)
            # Test has_header = false
            response = await family.import_families(
                testFile,
                has_header=False,
                delimiter='\t',
                connection=self.connection,
            )
            self.assertDictEqual(response, {'success': True})



from db.python.layers.participant import ParticipantLayer
from models.models import (
    PRIMARY_EXTERNAL_ORG,
    AssayUpsertInternal,
    ParticipantUpsertInternal,
    SampleUpsertInternal,
    SequencingGroupUpsertInternal,
)
from test.testbase import DbIsolatedTest, run_as_sync

default_assay_meta = {
    'sequencing_type': 'genome',
    'sequencing_technology': 'short-read',
    'sequencing_platform': 'illumina',
}

all_participants = [
    ParticipantUpsertInternal(
        external_ids={PRIMARY_EXTERNAL_ORG: 'Demeter'},
        meta={},
        samples=[
            SampleUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'sample_id001'},
                meta={},
                sequencing_groups=[
                    SequencingGroupUpsertInternal(
                        type='genome',
                        technology='short-read',
                        platform='illumina',
                        meta={},
                        assays=[
                            AssayUpsertInternal(
                                meta={
                                    'reads': [
                                        {
                                            'basename': 'sample_id001.filename-R1.fastq.gz',
                                            'checksum': None,
                                            'class': 'File',
                                            'location': '/path/to/sample_id001.filename-R1.fastq.gz',
                                            'size': 111,
                                        },
                                        {
                                            'basename': 'sample_id001.filename-R2.fastq.gz',
                                            'checksum': None,
                                            'class': 'File',
                                            'location': '/path/to/sample_id001.filename-R2.fastq.gz',
                                            'size': 111,
                                        },
                                    ],
                                    'reads_type': 'fastq',
                                    **default_assay_meta,
                                },
                                type='sequencing',
                            )
                        ],
                    ),
                    SequencingGroupUpsertInternal(
                        type='exome',
                        technology='short-read',
                        platform='illumina',
                        meta={},
                        assays=[
                            AssayUpsertInternal(
                                meta={
                                    'reads': [
                                        {
                                            'basename': 'sample_id001.exome.filename-R1.fastq.gz',
                                            'checksum': None,
                                            'class': 'File',
                                            'location': '/path/to/sample_id001.exome.filename-R1.fastq.gz',
                                            'size': 111,
                                        },
                                        {
                                            'basename': 'sample_id001.exome.filename-R2.fastq.gz',
                                            'checksum': None,
                                            'class': 'File',
                                            'location': '/path/to/sample_id001.exome.filename-R2.fastq.gz',
                                            'size': 111,
                                        },
                                    ],
                                    'reads_type': 'fastq',
                                    **default_assay_meta,
                                },
                                type='sequencing',
                            )
                        ],
                    ),
                ],
                type='blood',
            )
        ],
    ),
    ParticipantUpsertInternal(
        external_ids={PRIMARY_EXTERNAL_ORG: 'Apollo'},
        meta={},
        samples=[
            SampleUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'sample_id002'},
                meta={},
                sequencing_groups=[
                    SequencingGroupUpsertInternal(
                        type='genome',
                        technology='short-read',
                        platform='illumina',
                        meta={},
                        assays=[
                            AssayUpsertInternal(
                                meta={
                                    'reads': [
                                        {
                                            'basename': 'sample_id002.filename-R1.fastq.gz',
                                            'checksum': None,
                                            'class': 'File',
                                            'location': '/path/to/sample_id002.filename-R1.fastq.gz',
                                            'size': 111,
                                        },
                                        {
                                            'basename': 'sample_id002.filename-R2.fastq.gz',
                                            'checksum': None,
                                            'class': 'File',
                                            'location': '/path/to/sample_id002.filename-R2.fastq.gz',
                                            'size': 111,
                                        },
                                    ],
                                    'reads_type': 'fastq',
                                    **default_assay_meta,
                                },
                                type='sequencing',
                            )
                        ],
                    ),
                ],
                type='blood',
            ),
            SampleUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'sample_id004'},
                meta={},
                sequencing_groups=[
                    SequencingGroupUpsertInternal(
                        type='genome',
                        technology='short-read',
                        platform='illumina',
                        meta={},
                        assays=[
                            AssayUpsertInternal(
                                meta={
                                    'reads': [
                                        {
                                            'basename': 'sample_id004.filename-R1.fastq.gz',
                                            'checksum': None,
                                            'class': 'File',
                                            'location': '/path/to/sample_id004.filename-R1.fastq.gz',
                                            'size': 111,
                                        },
                                        {
                                            'basename': 'sample_id004.filename-R2.fastq.gz',
                                            'checksum': None,
                                            'class': 'File',
                                            'location': '/path/to/sample_id004.filename-R2.fastq.gz',
                                            'size': 111,
                                        },
                                    ],
                                    'reads_type': 'fastq',
                                    **default_assay_meta,
                                },
                                type='sequencing',
                            )
                        ],
                    )
                ],
                type='blood',
            ),
        ],
    ),
    ParticipantUpsertInternal(
        external_ids={PRIMARY_EXTERNAL_ORG: 'Athena'},
        meta={},
        samples=[
            SampleUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'sample_id003'},
                meta={},
                sequencing_groups=[
                    SequencingGroupUpsertInternal(
                        type='genome',
                        technology='short-read',
                        platform='illumina',
                        meta={},
                        assays=[],
                    )
                ],
                type='blood',
            )
        ],
    ),
]


class TestUpsert(DbIsolatedTest):
    """
    Test upsert functionality in SM
    """

    @run_as_sync
    async def test_insert_participants(self):
        """
        Test inserting participants, samples and sequences, and make sure they're correctly linked.

        Tests the other side of:
            tests.test_parse_generic_metadata:TestParseGenericMetadata.test_rows_with_participants
        """

        # Table interfaces
        pt = ParticipantLayer(self.connection)

        await pt.upsert_participants(all_participants, open_transaction=False)

        expected_sample_eid_to_participant_eid = {
            sample_eid: participant_eid
            for participant in all_participants
            for participant_eid in participant.external_ids.values()
            for sample in participant.samples
            for sample_eid in sample.external_ids.values()
        }

        db_participants = await self.connection.connection.fetch_all(
            'SELECT * FROM participant_external_id ORDER BY participant_id'
        )
        self.assertEqual(3, len(db_participants))
        self.assertEqual('Demeter', db_participants[0]['external_id'])
        self.assertEqual('Apollo', db_participants[1]['external_id'])
        self.assertEqual('Athena', db_participants[2]['external_id'])

        participant_id_map = {
            p['external_id']: p['participant_id'] for p in db_participants
        }

        db_samples = await self.connection.connection.fetch_all(
            """
            SELECT s.participant_id, seid.external_id
            FROM sample s
            INNER JOIN sample_external_id seid ON s.id = seid.sample_id
            WHERE seid.name = :PRIMARY_EXTERNAL_ORG
            ORDER BY s.id
            """,
            {'PRIMARY_EXTERNAL_ORG': PRIMARY_EXTERNAL_ORG},
        )
        self.assertEqual(4, len(db_samples))
        for db_sample in db_samples:
            self.assertIsNotNone(db_sample['external_id'])
            self.assertIsNotNone(db_sample['participant_id'])
            # get expected_participant_id from the db_sample external_id
            expected_participant_eid = expected_sample_eid_to_participant_eid.get(
                db_sample['external_id']
            )
            self.assertEqual(
                participant_id_map[expected_participant_eid],
                db_sample['participant_id'],
            )

        db_sequencing_groups = await self.connection.connection.fetch_all(
            'SELECT * FROM sequencing_group'
        )
        self.assertEqual(5, len(db_sequencing_groups))
        for db_sg in db_sequencing_groups:
            self.assertIsNotNone(db_sg['sample_id'])
            self.assertIsNotNone(db_sg['type'])

        db_assays = await self.connection.connection.fetch_all('SELECT * FROM assay')

        self.assertEqual(4, len(db_assays))
        for db_a in db_assays:
            self.assertIsNotNone(db_a['sample_id'])
            self.assertIsNotNone(db_a['type'])

        db_participant_no_assays = await self.connection.connection.fetch_one(
            """
            SELECT COUNT(DISTINCT a.id) AS cnt
            FROM sample AS s
            INNER JOIN participant AS p ON p.id = s.participant_id
            INNER JOIN participant_external_id AS pei ON p.id = pei.participant_id
            LEFT JOIN assay AS a ON a.sample_id = s.id
            WHERE pei.external_id = "Athena"
            """
        )

        self.assertEqual(0, db_participant_no_assays['cnt'])

        db_participant_has_assays = await self.connection.connection.fetch_one(
            """
            SELECT COUNT(DISTINCT a.id) AS cnt
            FROM sample AS s
            INNER JOIN participant AS p ON p.id = s.participant_id
            INNER JOIN participant_external_id AS pei ON p.id = pei.participant_id
            LEFT JOIN assay AS a ON a.sample_id = s.id
            WHERE pei.external_id = "Apollo"
            """
        )

        self.assertEqual(2, db_participant_has_assays['cnt'])



# pylint: disable=protected-access too-many-public-methods
from datetime import datetime
from typing import Any
from unittest import mock

import google.cloud.bigquery as bq

from db.python.tables.bq.billing_base import (
    BillingBaseTable,
)
from db.python.tables.bq.billing_daily_extended import BillingDailyExtendedTable
from db.python.tables.bq.billing_filter import BillingFilter
from db.python.tables.bq.billing_utils import (
    abbrev_cost_category,
    append_detailed_cost_records,
    append_total_running_cost,
    convert_output,
    filter_to_optimise_query,
    last_loaded_day_filter,
    prepare_aggregation,
    prepare_order_by_string,
    prepare_time_periods,
)
from db.python.tables.bq.generic_bq_filter import GenericBQFilter
from models.enums import BillingTimePeriods
from models.models import (
    BillingColumn,
    BillingCostBudgetRecord,
    BillingCostDetailsRecord,
    BillingRunningCostQueryModel,
    BillingTotalCostQueryModel,
)
from test.testbase import run_as_sync
from test.testbqbase import BqTest


def mock_execute_query_running_cost(query, *_args, **_kwargs):
    """
    This is a mockup function for _execute_query function
    This returns one mockup BQ query result
    for 2 different SQL queries used by get_running_cost API point
    Those 2 queries are:
    1. query to get last loaded day
    2. query to get aggregated monthly/daily cost
    """
    if ' as last_loaded_day' in query:
        # This is the 1st query to get last loaded day
        # mockup BQ query result for last_loaded_day as list of rows
        return [
            mock.MagicMock(spec=bq.Row, last_loaded_day='2024-01-01 00:00:00+00:00')
        ]

    # This is the 2nd query to get aggregated monthly cost
    # return mockup BQ query result as list of rows
    return [
        mock.MagicMock(
            spec=bq.Row,
            field='TOPIC1',
            cost_category='Compute Engine',
            daily_cost=123.45,
            monthly_cost=2345.67,
        )
    ]


def mock_execute_query_get_total_cost(_query, *_args, **_kwargs):
    """
    This is a mockup function for _execute_query function
    This returns one mockup BQ query result
    """
    # mockup BQ query result topic cost by invoice month, return row iterator
    mock_rows = mock.MagicMock(spec=bq.table.RowIterator)
    mock_rows.total_rows = 3
    mock_rows.__iter__.return_value = [
        {'day': '202301', 'topic': 'TOPIC1', 'cost': 123.10},
        {'day': '202302', 'topic': 'TOPIC1', 'cost': 223.20},
        {'day': '202303', 'topic': 'TOPIC1', 'cost': 323.30},
    ]
    mock_result = mock.MagicMock(spec=bq.job.QueryJob)
    mock_result.result.return_value = mock_rows
    return mock_result


class TestBillingBaseTable(BqTest):
    """Test BillingBaseTable and its methods"""

    def setUp(self):
        super().setUp()

        # setup table object
        # base is abstract, so we need to use a child class
        # DailyExtended is the simplest one, almost no overrides
        self.table_obj = BillingDailyExtendedTable(self.connection)

    def test_query_to_partitioned_filter(self):
        """Test query to partitioned filter conversion"""

        # given
        start_date = '2023-01-01'
        end_date = '2024-01-01'
        filters: dict[BillingColumn, str | list[Any] | dict[Any, Any]] = {
            BillingColumn.TOPIC: 'TEST_TOPIC'
        }

        # expected
        expected_filter = BillingFilter(
            day=GenericBQFilter(
                gte=datetime(2023, 1, 1, 0, 0),
                lte=datetime(2024, 1, 1, 0, 0),
            ),
            topic=GenericBQFilter(eq='TEST_TOPIC'),
        )

        query = BillingTotalCostQueryModel(
            fields=[],  # not relevant for this test, but can't be null generally
            start_date=start_date,
            end_date=end_date,
            filters=filters,
        )
        filter_ = BillingBaseTable._query_to_partitioned_filter(query)

        # BillingFilter has __eq__ method, so we can compare them directly
        self.assertEqual(expected_filter, filter_)

    def test_abbrev_cost_category(self):
        """Test abbrev_cost_category"""

        # table name is set in the class
        categories_to_expected = {
            'Cloud Storage': 'S',
            'Compute Engine': 'C',
            'Other': 'C',
        }

        # test category to abreveation
        for cat, expected_abrev in categories_to_expected.items():
            self.assertEqual(expected_abrev, abbrev_cost_category(cat))

    def test_prepare_time_periods_by_day(self):
        """Test prepare_time_periods"""

        query = BillingTotalCostQueryModel(
            fields=[],
            start_date='2024-01-01',
            end_date='2024-01-01',
            time_periods=BillingTimePeriods.DAY,
        )

        time_group = prepare_time_periods(query)

        self.assertEqual('FORMAT_DATE("%Y-%m-%d", day) as day', time_group.field)
        self.assertEqual('PARSE_DATE("%Y-%m-%d", day) as day', time_group.formula)
        self.assertEqual(',', time_group.separator)

    def test_prepare_time_periods_by_week(self):
        """Test prepare_time_periods"""

        query = BillingTotalCostQueryModel(
            fields=[],
            start_date='2024-01-01',
            end_date='2024-01-01',
            time_periods=BillingTimePeriods.WEEK,
        )

        time_group = prepare_time_periods(query)

        self.assertEqual('FORMAT_DATE("%Y%W", day) as day', time_group.field)
        self.assertEqual('PARSE_DATE("%Y%W", day) as day', time_group.formula)
        self.assertEqual(',', time_group.separator)

    def test_prepare_time_periods_by_month(self):
        """Test prepare_time_periods"""

        query = BillingTotalCostQueryModel(
            fields=[],
            start_date='2024-01-01',
            end_date='2024-01-01',
            time_periods=BillingTimePeriods.MONTH,
        )

        time_group = prepare_time_periods(query)

        self.assertEqual('FORMAT_DATE("%Y%m", day) as day', time_group.field)
        self.assertEqual('PARSE_DATE("%Y%m", day) as day', time_group.formula)
        self.assertEqual(',', time_group.separator)

    def test_prepare_time_periods_by_invoice_month(self):
        """Test prepare_time_periods"""

        query = BillingTotalCostQueryModel(
            fields=[],
            start_date='2024-01-01',
            end_date='2024-01-01',
            time_periods=BillingTimePeriods.INVOICE_MONTH,
        )

        time_group = prepare_time_periods(query)

        self.assertEqual('invoice_month as day', time_group.field)
        self.assertEqual('PARSE_DATE("%Y%m", day) as day', time_group.formula)
        self.assertEqual(',', time_group.separator)

    def test_filter_to_optimise_query(self):
        """Test _filter_to_optimise_query"""

        result = filter_to_optimise_query()
        self.assertEqual(
            'day >= TIMESTAMP(@start_day) AND day <= TIMESTAMP(@last_day)', result
        )

    def test_last_loaded_day_filter(self):
        """Test _last_loaded_day_filter"""

        result = last_loaded_day_filter()
        self.assertEqual('day = TIMESTAMP(@last_loaded_day)', result)

    def test_convert_output_empty_results(self):
        """Test _convert_output - various empty results"""

        empty_results = convert_output(None)
        self.assertEqual([], empty_results)

        query_job_result = mock.MagicMock(spec=bq.job.QueryJob)
        query_job_result.result.total_rows = 0

        empty_list = convert_output(query_job_result)
        self.assertEqual([], empty_list)

        query_job_result = mock.MagicMock(spec=bq.job.QueryJob)
        query_job_result.result.return_value = mock.MagicMock(spec=bq.table.RowIterator)

        empty_row_iterator = convert_output(query_job_result)
        self.assertEqual([], empty_row_iterator)

    def test_convert_output_one_record(self):
        """Test _convert_output - one record result"""

        mock_rows = mock.MagicMock(spec=bq.table.RowIterator)
        mock_rows.total_rows = 1
        mock_rows.__iter__.return_value = [{}]

        query_job_result = mock.MagicMock(spec=bq.job.QueryJob)
        query_job_result.result.return_value = mock_rows

        single_row = convert_output(query_job_result)
        self.assertEqual([{}], single_row)

    def test_convert_output_label_record(self):
        """Test _convert_output - test with label item"""
        mock_rows = mock.MagicMock(spec=bq.table.RowIterator)
        mock_rows.total_rows = 1
        mock_rows.__iter__.return_value = [
            {'labels': [{'key': 'test_key', 'value': 'test_value'}]}
        ]

        query_job_result = mock.MagicMock(spec=bq.job.QueryJob)
        query_job_result.result.return_value = mock_rows

        row_iterator = convert_output(query_job_result)
        self.assertEqual(
            [
                {
                    # keep the original labels
                    'labels': [{'key': 'test_key', 'value': 'test_value'}],
                    # append the labels as key-value pairs
                    'test_key': 'test_value',
                }
            ],
            row_iterator,
        )

    def test_prepare_order_by_string_empty(self):
        """Test _prepare_order_by_string - empty results"""

        self.assertEqual('', prepare_order_by_string(None))

    def test_prepare_order_by_string_order_by_one_column(self):
        """Test _prepare_order_by_string"""

        # DESC order by column
        self.assertEqual(
            'ORDER BY cost DESC',
            prepare_order_by_string({BillingColumn.COST: True}),
        )

        # ASC order by column
        self.assertEqual(
            'ORDER BY cost ASC',
            prepare_order_by_string({BillingColumn.COST: False}),
        )

    def test_prepare_order_by_string_order_by_two_columns(self):
        """Test _prepare_order_by_string - order by 2 columns"""
        self.assertEqual(
            'ORDER BY cost ASC,day DESC',
            prepare_order_by_string(
                {BillingColumn.COST: False, BillingColumn.DAY: True}
            ),
        )

    def test_prepare_aggregation_default_group_by(self):
        """Test _prepare_aggregation"""

        query = BillingTotalCostQueryModel(
            fields=[], start_date='2024-01-01', end_date='2024-01-01'
        )

        fields_selected, group_by = prepare_aggregation(query)
        # no fields selected so it is empty
        self.assertEqual('', fields_selected)
        # by default results are grouped by day
        self.assertEqual('GROUP BY day', group_by)

    def test_prepare_aggregation_default_no_grouping_by(self):
        """Test _prepare_aggregation"""

        # test when query is not grouped by
        query = BillingTotalCostQueryModel(
            fields=[BillingColumn.TOPIC],
            start_date='2024-01-01',
            end_date='2024-01-01',
            group_by=False,
        )

        fields_selected, group_by = prepare_aggregation(query)
        # topic field is selected
        self.assertEqual('topic', fields_selected)
        # group by is switched off
        self.assertEqual('', group_by)

    def test_prepare_aggregation_default_group_by_more_columns(self):
        """Test _prepare_aggregation"""

        # test when query is grouped by, but column can not be grouped by
        # cost can not be grouped by, so it is not present in the result
        query = BillingTotalCostQueryModel(
            fields=[BillingColumn.TOPIC, BillingColumn.COST],
            start_date='2024-01-01',
            end_date='2024-01-01',
            group_by=True,
        )

        fields_selected, group_by = prepare_aggregation(query)
        self.assertEqual('topic', fields_selected)
        # always group by day and any field that can be grouped by
        self.assertEqual('GROUP BY day,topic', group_by)

    def test_execute_query_results_as_list(self):
        """Test _execute_query"""

        # we are not running SQL against real BQ, just a mocking, so we can use any query
        sql_query = 'SELECT 1;'
        sql_params: list[Any] = []

        # test results_as_list=True
        given_bq_results = [[], [123], ['a', 'b', 'c']]
        for bq_result in given_bq_results:
            self.bq_result.result.return_value = bq_result
            results = self.table_obj._execute_query(
                sql_query, sql_params, results_as_list=True
            )
            self.assertEqual(bq_result, results)

    def test_execute_query_results_not_as_list(self):
        """Test _execute_query"""

        # we are not running SQL against real BQ, just a mocking, so we can use any query
        sql_query = 'SELECT 1;'
        sql_params: list[Any] = []

        # now test results_as_list=False
        given_bq_results = [[], [123], ['a', 'b', 'c']]
        for bq_result in given_bq_results:
            # mock BigQuery result
            self.bq_result.result.return_value = bq_result
            self.bq_result.total_bytes_processed = 0
            results = self.table_obj._execute_query(
                sql_query, sql_params, results_as_list=True
            )
            self.assertEqual(bq_result, results)

    def test_execute_query_with_sql_params(self):
        """Test _execute_query"""

        # now test results_as_list=False and with some dummy params
        sql_query = 'SELECT 1;'
        sql_params = [
            bq.ScalarQueryParameter('dummy_not_used', 'STRING', '2021-01-01 00:00:00')
        ]

        given_bq_results = [[], [123], ['a', 'b', 'c']]
        for bq_result in given_bq_results:
            # mock BigQuery result
            self.bq_result.result.return_value = bq_result
            self.bq_result.total_bytes_processed = 0
            results = self.table_obj._execute_query(
                sql_query, sql_params, results_as_list=True
            )
            self.assertEqual(bq_result, results)

    @run_as_sync
    async def test_append_total_running_cost_no_topic(self):
        """Test _append_total_running_cost"""

        # test _append_total_running_cost function, no topic present
        total_record = await append_total_running_cost(
            field=BillingColumn.TOPIC,
            is_current_month=True,
            last_loaded_day=None,
            total_monthly={'C': {'ALL': 1000}, 'S': {'ALL': 2000}},
            total_daily={'C': {'ALL': 100}, 'S': {'ALL': 200}},
            total_monthly_category={},
            total_daily_category={},
            results=[],
        )

        self.assertEqual(
            [
                BillingCostBudgetRecord(
                    field='All Topics',
                    total_monthly=3000.0,
                    total_daily=300.0,
                    compute_monthly=1000.0,
                    compute_daily=100.0,
                    storage_monthly=2000.0,
                    storage_daily=200.0,
                    details=[],
                    budget_spent=None,
                    budget=None,
                    last_loaded_day=None,
                )
            ],
            total_record,
        )

    @run_as_sync
    async def test_append_total_running_cost_not_current_month(self):
        """Test _append_total_running_cost"""

        # test _append_total_running_cost function, not current month
        total_record = await append_total_running_cost(
            field=BillingColumn.TOPIC,
            is_current_month=False,
            last_loaded_day=None,
            total_monthly={'C': {'ALL': 1000}, 'S': {'ALL': 2000}},
            total_daily=None,
            total_monthly_category={},
            total_daily_category={},
            results=[],
        )

        self.assertEqual(
            [
                BillingCostBudgetRecord(
                    field='All Topics',
                    total_monthly=3000.0,
                    total_daily=None,
                    compute_monthly=1000.0,
                    compute_daily=None,
                    storage_monthly=2000.0,
                    storage_daily=None,
                    details=[],
                    budget_spent=None,
                    budget=None,
                    last_loaded_day=None,
                )
            ],
            total_record,
        )

    @run_as_sync
    async def test_append_total_running_cost_current_month(self):
        """Test _append_total_running_cost"""

        total_record = await append_total_running_cost(
            field=BillingColumn.TOPIC,
            is_current_month=True,
            last_loaded_day=None,
            total_monthly={'C': {'ALL': 1000}, 'S': {'ALL': 2000}},
            total_daily={'C': {'ALL': 100}, 'S': {'ALL': 200}},
            total_monthly_category={
                'Compute Engine': 900,
                'Cloud Storage': 2000,
                'Other': 100,
            },
            total_daily_category={
                'Compute Engine': 90,
                'Cloud Storage': 200,
                'Other': 10,
            },
            results=[],
        )

        self.assertEqual(
            [
                BillingCostBudgetRecord(
                    field='All Topics',
                    total_monthly=3000.0,
                    total_daily=300.0,
                    compute_monthly=1000.0,
                    compute_daily=100.0,
                    storage_monthly=2000.0,
                    storage_daily=200.0,
                    details=[
                        BillingCostDetailsRecord(
                            cost_group='C',
                            cost_category='Compute Engine',
                            daily_cost=90.0,
                            monthly_cost=900.0,
                        ),
                        BillingCostDetailsRecord(
                            cost_group='S',
                            cost_category='Cloud Storage',
                            daily_cost=200.0,
                            monthly_cost=2000.0,
                        ),
                        BillingCostDetailsRecord(
                            cost_group='C',
                            cost_category='Other',
                            daily_cost=10.0,
                            monthly_cost=100.0,
                        ),
                    ],
                    budget_spent=None,
                    budget=None,
                    last_loaded_day=None,
                )
            ],
            total_record,
        )

    @run_as_sync
    async def test_budgets_by_gcp_project_empty_results(self):
        """Test _budgets_by_gcp_project"""

        # Only GCP_PROJECT and current month has budget
        empty_result = await self.table_obj._budgets_by_gcp_project(
            BillingColumn.TOPIC, False
        )
        self.assertEqual({}, empty_result)

        # GCP_PROJECT and current month, but BQ mockup setup as empty values
        empty_result = await self.table_obj._budgets_by_gcp_project(
            BillingColumn.GCP_PROJECT, True
        )
        self.assertEqual({}, empty_result)

    @run_as_sync
    async def test_budgets_by_gcp_project_with_results(self):
        """Test _budgets_by_gcp_project"""

        # GCP_PROJECT and current month and Mockup set as 2 records
        self.bq_result.result.return_value = [
            mock.MagicMock(spec=bq.Row, gcp_project='Project1', budget=1000.0),
            mock.MagicMock(spec=bq.Row, gcp_project='Project2', budget=2000.0),
        ]

        non_empty_result = await self.table_obj._budgets_by_gcp_project(
            BillingColumn.GCP_PROJECT, True
        )
        self.assertDictEqual({'Project1': 1000.0, 'Project2': 2000.0}, non_empty_result)

    @run_as_sync
    async def test_execute_running_cost_query_invalid_months(self):
        """Test _execute_running_cost_query"""

        # test invalid inputs
        with self.assertRaises(ValueError) as context:
            await self.table_obj._execute_running_cost_query_with_filters(
                BillingRunningCostQueryModel(
                    field=BillingColumn.TOPIC, invoice_month=None
                )
            )

        self.assertTrue('Invalid invoice month' in str(context.exception))

        with self.assertRaises(ValueError) as context:
            await self.table_obj._execute_running_cost_query_with_filters(
                BillingRunningCostQueryModel(
                    field=BillingColumn.TOPIC, invoice_month='12345678'
                )
            )

        self.assertTrue('Invalid invoice month' in str(context.exception))

        with self.assertRaises(ValueError) as context:
            await self.table_obj._execute_running_cost_query_with_filters(
                BillingRunningCostQueryModel(
                    field=BillingColumn.TOPIC, invoice_month='1024AA'
                )
            )
        self.assertTrue('Invalid invoice month' in str(context.exception))

    @run_as_sync
    async def test_execute_running_cost_query_empty_results_old_month(self):
        """Test _execute_running_cost_query"""

        # no mocked BQ results, should return as empty
        (
            is_current_month,
            last_loaded_day,
            query_job_result,
        ) = await self.table_obj._execute_running_cost_query_with_filters(
            BillingRunningCostQueryModel(
                field=BillingColumn.TOPIC, invoice_month='202101'
            )
        )

        self.assertEqual(False, is_current_month)
        self.assertEqual(None, last_loaded_day)
        self.assertEqual([], query_job_result)

    @run_as_sync
    async def test_execute_running_cost_query_empty_results_current_month(self):
        """Test _execute_running_cost_query"""

        # no mocked BQ results, should return as empty
        # use current month to test the current month branch
        current_month_as_string = datetime.now().strftime('%Y%m')
        (
            is_current_month,
            last_loaded_day,
            query_job_result,
        ) = await self.table_obj._execute_running_cost_query_with_filters(
            BillingRunningCostQueryModel(
                field=BillingColumn.TOPIC, invoice_month=current_month_as_string
            )
        )

        self.assertEqual(True, is_current_month)
        self.assertEqual(None, last_loaded_day)
        self.assertEqual([], query_job_result)

    @run_as_sync
    async def test_append_total_running_cost_empty_results(self):
        """Test append_total_running_cost"""

        # test empty results
        empty_results = await append_detailed_cost_records(
            budgets_per_gcp_project={},
            is_current_month=False,
            last_loaded_day=None,
            total_monthly={},
            total_daily={},
            field_details={},
            results=[],
        )

        self.assertEqual([], empty_results)

    @run_as_sync
    async def test_append_running_cost_records_simple_data(self):
        """Test append_running_cost_records"""

        # prepare simple input data
        field_details: dict[str, Any] = {
            'Project1': [],
        }

        simple_result = await append_detailed_cost_records(
            budgets_per_gcp_project={},
            is_current_month=False,
            last_loaded_day=None,
            total_monthly={'C': {}, 'S': {}},
            total_daily={'C': {}, 'S': {}},
            field_details=field_details,
            results=[],
        )

        self.assertEqual(
            [
                BillingCostBudgetRecord(
                    field='Project1',
                    total_monthly=0.0,
                    compute_monthly=0.0,
                    compute_daily=0.0,
                    storage_monthly=0.0,
                    storage_daily=0.0,
                    details=[],
                    last_loaded_day=None,
                    total_daily=None,
                    budget_spent=None,
                    budget=None,
                )
            ],
            simple_result,
        )

    @run_as_sync
    async def test_append_running_cost_records_with_details(self):
        """Test append_running_cost_records"""

        # prepare input data with more details
        field_details = {
            'Project2': [
                {
                    'cost_group': 'C',
                    'cost_category': 'Compute Engine',
                    'daily_cost': 90.0,
                    'monthly_cost': 900.0,
                }
            ],
        }

        detailed_result = await append_detailed_cost_records(
            budgets_per_gcp_project={},
            is_current_month=False,
            last_loaded_day=None,
            total_monthly={'C': {}, 'S': {}},
            total_daily={'C': {}, 'S': {}},
            field_details=field_details,
            results=[],
        )

        self.assertEqual(
            [
                BillingCostBudgetRecord(
                    field='Project2',
                    total_monthly=0.0,
                    compute_monthly=0.0,
                    compute_daily=0.0,
                    storage_monthly=0.0,
                    storage_daily=0.0,
                    details=[
                        BillingCostDetailsRecord(
                            cost_group='C',
                            cost_category='Compute Engine',
                            daily_cost=90.0,
                            monthly_cost=900.0,
                        )
                    ],
                    last_loaded_day=None,
                    total_daily=None,
                    budget_spent=None,
                    budget=None,
                )
            ],
            detailed_result,
        )

    @run_as_sync
    async def test_get_running_cost_invalid_input(self):
        """Test get_running_cost"""

        # test invalid outputs
        with self.assertRaises(ValueError) as context:
            await self.table_obj.get_running_cost_with_filters(
                # not allowed field
                BillingRunningCostQueryModel(
                    field=BillingColumn.SKU,
                    invoice_month=None,
                )
            )

        self.assertTrue(
            (
                'Invalid field only topic, dataset, gcp-project, compute_category, '
                'wdl_task_name, cromwell_sub_workflow_name & namespace are allowed'
            )
            in str(context.exception)
        )

    @run_as_sync
    async def test_get_running_cost_empty_results(self):
        """Test get_running_cost"""

        # test empty cost (no BQ mockup data provided)
        empty_results = await self.table_obj.get_running_cost_with_filters(
            BillingRunningCostQueryModel(
                field=BillingColumn.TOPIC,
                invoice_month='202301',
            )
        )

        self.assertEqual([], empty_results)

    @run_as_sync
    async def test_get_running_cost_older_month(self):
        """Test get_running_cost"""

        # mockup BQ sql query result for _execute_running_cost_query function
        self.table_obj._execute_query = mock.MagicMock(
            side_effect=mock_execute_query_running_cost
        )

        one_record_result = await self.table_obj.get_running_cost_with_filters(
            BillingRunningCostQueryModel(
                field=BillingColumn.TOPIC,
                invoice_month='202301',
            )
        )

        self.assertEqual(
            [
                BillingCostBudgetRecord(
                    field='All Topics',
                    total_monthly=2345.67,
                    total_daily=None,
                    compute_monthly=2345.67,
                    compute_daily=None,
                    storage_monthly=0.0,
                    storage_daily=None,
                    details=[
                        BillingCostDetailsRecord(
                            cost_group='C',
                            cost_category='Compute Engine',
                            daily_cost=None,
                            monthly_cost=2345.67,
                        )
                    ],
                    budget_spent=None,
                    budget=None,
                    last_loaded_day=None,
                ),
                BillingCostBudgetRecord(
                    field='TOPIC1',
                    total_monthly=2345.67,
                    total_daily=None,
                    compute_monthly=2345.67,
                    compute_daily=0.0,
                    storage_monthly=0.0,
                    storage_daily=0.0,
                    details=[
                        BillingCostDetailsRecord(
                            cost_group='C',
                            cost_category='Compute Engine',
                            daily_cost=None,
                            monthly_cost=2345.67,
                        )
                    ],
                    budget_spent=None,
                    budget=None,
                    last_loaded_day=None,
                ),
            ],
            one_record_result,
        )

    @run_as_sync
    async def test_get_running_cost_current_month(self):
        """Test get_running_cost"""

        # mockup BQ sql query result for _execute_running_cost_query function
        self.table_obj._execute_query = mock.MagicMock(
            side_effect=mock_execute_query_running_cost
        )

        # use the current month to test the current month branch
        current_month_as_string = datetime.now().strftime('%Y%m')

        current_month_result = await self.table_obj.get_running_cost_with_filters(
            BillingRunningCostQueryModel(
                field=BillingColumn.TOPIC,
                invoice_month=current_month_as_string,
            )
        )

        self.assertEqual(
            [
                BillingCostBudgetRecord(
                    field='All Topics',
                    total_monthly=2345.67,
                    total_daily=123.45,
                    compute_monthly=2345.67,
                    compute_daily=123.45,
                    storage_monthly=0.0,
                    storage_daily=0.0,
                    details=[
                        BillingCostDetailsRecord(
                            cost_group='C',
                            cost_category='Compute Engine',
                            daily_cost=123.45,
                            monthly_cost=2345.67,
                        )
                    ],
                    budget_spent=None,
                    budget=None,
                    last_loaded_day='Jan 01',
                ),
                BillingCostBudgetRecord(
                    field='TOPIC1',
                    total_monthly=2345.67,
                    total_daily=123.45,
                    compute_monthly=2345.67,
                    compute_daily=123.45,
                    storage_monthly=0.0,
                    storage_daily=0.0,
                    details=[
                        BillingCostDetailsRecord(
                            cost_group='C',
                            cost_category='Compute Engine',
                            daily_cost=123.45,
                            monthly_cost=2345.67,
                        )
                    ],
                    budget_spent=None,
                    budget=None,
                    last_loaded_day='Jan 01',
                ),
            ],
            current_month_result,
        )

    @run_as_sync
    async def test_get_total_cost(self):
        """Test get_total_cost"""

        # test invalid input
        query = BillingTotalCostQueryModel(
            fields=[], start_date='2023-01-01', end_date='2024-01-01'
        )

        with self.assertRaises(ValueError) as context:
            await self.table_obj.get_total_cost(query)

        self.assertTrue('Date and Fields are required' in str(context.exception))

        # test empty results
        query = BillingTotalCostQueryModel(
            fields=[BillingColumn.TOPIC],
            start_date='2023-01-01',
            end_date='2024-01-01',
            time_periods=BillingTimePeriods.INVOICE_MONTH,
        )

        # no BQ mockup data setup, returns empty list
        empty_results = await self.table_obj.get_total_cost(query)
        self.assertEqual([], empty_results)

        # mockup BQ sql query result for _execute_query to return 3 records.
        # implementation is inside mock_execute_query function
        self.table_obj._execute_query = mock.MagicMock(
            side_effect=mock_execute_query_get_total_cost
        )
        results = await self.table_obj.get_total_cost(query)
        self.assertEqual(
            [
                {'day': '202301', 'topic': 'TOPIC1', 'cost': 123.1},
                {'day': '202302', 'topic': 'TOPIC1', 'cost': 223.2},
                {'day': '202303', 'topic': 'TOPIC1', 'cost': 323.3},
            ],
            results,
        )



# pylint: disable=no-member
import logging
import os
from collections import defaultdict
from datetime import datetime
from typing import Callable, Iterable, TypeVar

from cloudpathlib import AnyPath, GSPath
from google.cloud import storage

# type declarations for GroupBy
T = TypeVar('T')
X = TypeVar('X')


def group_by(iterable: Iterable[T], selector: Callable[[T], X]) -> dict[X, list[T]]:
    """Simple group by implementation"""
    ret: dict[X, list[T]] = defaultdict(list)
    for k in iterable:
        ret[selector(k)].append(k)

    return dict(ret)


class CloudHelper:
    """General CloudHelper for parsers"""

    LOCAL_PREFIX = '/'
    GCS_PREFIX = 'gs://'
    AZ_PREFIX = 'az://'

    def __init__(self, search_paths: list[str] | None):
        # gs specific
        self._gcs_client = None
        self.gcs_bucket_refs: dict[str, storage.Bucket] = {}

        self.search_paths = search_paths or []

        self.filename_map: dict[str, str] = self.populate_filename_map(
            self.search_paths
        )
        # pylint: disable

    @staticmethod
    def guess_delimiter_from_filename(filename: str):
        """
        Guess delimiter from filename
        """
        extension_to_delimiter = {'.csv': ',', '.tsv': '\t'}
        relevant_delimiter = next(
            (
                delimiter
                for ext, delimiter in extension_to_delimiter.items()
                if filename.endswith(ext)
            ),
            None,
        )
        if relevant_delimiter:
            return relevant_delimiter

        raise ValueError(f'Unrecognised extension on file: {filename}')

    def file_path(self, filename: str, raise_exception: bool = True) -> str | None:
        """
        If the filename is a fully qualified path, this does NOT check for file existence
        """
        if any(
            filename.startswith(prefix)
            for prefix in (self.GCS_PREFIX, '/', self.AZ_PREFIX)
        ):
            return filename

        expanded_local_path = os.path.abspath(filename)
        if expanded_local_path.startswith('.'):
            expanded_local_path = os.path.abspath(filename)
        if os.path.exists(expanded_local_path):
            return expanded_local_path

        if not self.filename_map:
            raise AssertionError(
                'If no search path is provided, the file_path must be overridden'
            )

        if filename not in self.filename_map:
            if raise_exception:
                raise FileNotFoundError(
                    f'Could not find file {filename!r} in search paths'
                )
            return None

        return self.filename_map[filename]

    def list_directory(self, directory_name) -> list[str]:
        """List directory"""
        path = self.file_path(directory_name)

        # this was super slow to perform, do don't do this atm
        # return [os.path.join(path, p.name) for p in AnyPath(path).iterdir()]

        if path.startswith('gs://'):
            return self._list_gcs_directory(path)

        if path.startswith('/'):
            return [os.path.join(path, f) for f in os.listdir(path)]

        raise ValueError(f'Could not handle listing directory of {directory_name!r}')

    async def file_contents(self, filename) -> str | None:
        """Get contents of file (decoded as utf8)"""
        path = self.file_path(filename)
        with AnyPath(path).open(encoding='utf-8-sig') as f:
            return f.read()

    async def file_exists(self, filename: str) -> bool:
        """Determines whether a file exists"""
        path = self.file_path(filename)
        if path.startswith('gs://'):
            # add a specific case for GCS as the AnyPath implementation calls
            # bucket.get_blob which triggers a read (which humans are not permitted)
            blob = await self.get_gcs_blob(path)
            return blob is not None

        return AnyPath(path).exists()

    async def file_size(self, filename) -> int:
        """Get size of file in bytes"""
        return AnyPath(self.file_path(filename)).stat().st_size

    async def datetime_added(self, filename) -> datetime | None:
        """Get the date and time the file was created"""
        path = self.file_path(filename)
        if path.startswith('gs://'):
            # add a specific case for GCS as the AnyPath implementation calls
            # bucket.get_blob which triggers a read (which humans are not permitted)
            blob = await self.get_gcs_blob(path)
            return blob.time_created

        ctime = AnyPath(path).stat().st_ctime
        if not ctime:
            return None

        return datetime.utcfromtimestamp(ctime)

    def populate_filename_map(self, search_locations: list[str]) -> dict[str, str]:
        """
        FileMapParser uses search locations based on the filename,
        so let's prepopulate that filename_map from the search_locations!
        """

        fn_map: dict[str, str] = {}
        for directory in search_locations:
            directory_list = self.list_directory(directory)
            for file in directory_list:
                file = file.strip()
                file_base = os.path.basename(file)
                if file_base in fn_map:
                    logging.warning(
                        f'File {file!r} from {directory!r} already exists in directory map: {fn_map[file_base]}'
                    )
                    continue
                fn_map[file_base] = file

        return fn_map

    # GCS specific methods
    @property
    def gcs_client(self) -> storage.Client:
        """Get GCP storage client"""
        if not self._gcs_client:
            self._gcs_client = storage.Client()
        return self._gcs_client

    def get_gcs_bucket(self, bucket_name: str) -> storage.Bucket:
        """Get cached bucket client from optional bucket name"""
        assert bucket_name
        if bucket_name not in self.gcs_bucket_refs:
            self.gcs_bucket_refs[bucket_name] = self.gcs_client.get_bucket(bucket_name)

        return self.gcs_bucket_refs[bucket_name]

    async def get_gcs_blob(self, filename: str) -> storage.Blob | None:
        """Convenience function for getting blob from fully qualified GCS path"""
        if not filename.startswith(self.GCS_PREFIX):
            raise ValueError('No blob available')

        blob = storage.Blob.from_string(filename, client=self.gcs_client)

        if not blob.exists():
            # maintain existing behaviour
            return None

        if not blob.time_created:
            # GCP blobs sometimes need a reload
            blob.reload()

        return blob

    def _list_gcs_directory(self, gcs_path) -> list[str]:
        path = GSPath(gcs_path)
        if path.parts[2:]:
            remaining_path = '/'.join(path.parts[2:]) + '/'  # has to end with "/"
        else:
            remaining_path = None
        blobs = self.gcs_client.list_blobs(
            path.bucket, prefix=remaining_path, delimiter='/'
        )
        return [f'gs://{path.bucket}/{blob.name}' for blob in blobs]



from unittest import mock
from test.testbase import DbIsolatedTest, run_as_sync
from datetime import date

from db.python.tables.sample import SampleTable


def mock_fetch_all_sample_count(fetch_all_return: list[dict[str, int]]):
    """
    This is a mockup function for fetch_all function on the Database object.
    This returns one mockup of a query result that contains the the number of samples
    in a project for each month.
    """

    def sample_count_mock(*_args, **_kwargs):
        mock_rows = mock.AsyncMock(args={}, spec=['__iter__', '__next__'])
        mock_rows.total_rows = len(fetch_all_return)
        mock_rows.__iter__.return_value = fetch_all_return

        return mock_rows

    return sample_count_mock


def mock_today(date_to_mock: date):
    """
    Returns a function decorator that patches the return value of date.today() before
    the given test_fn is called.
    Additionally, the fromisoformat of the date class is configured to function as it would
    without mocking.
    """

    def today_decorator(test_fn):
        async def test_fn_wrapper(self):
            patcher = mock.patch('db.python.tables.sample.date')
            mock_date = patcher.start()
            mock_date.today.return_value = date_to_mock
            # https://docs.python.org/3/library/unittest.mock-examples.html#partial-mocking
            mock_date.fromisoformat.side_effect = date.fromisoformat

            await test_fn(self)

            patcher.stop()

        return test_fn_wrapper

    return today_decorator


class TestMonthlySamplesPerProject(DbIsolatedTest):
    """Tests for retrieving the accumulated samples per month."""

    @run_as_sync
    async def setUp(self) -> None:
        super().setUp()

        self.sample_table = SampleTable(self.connection)

    @run_as_sync
    @mock_today(date(year=2025, month=9, day=1))
    async def test_standard_date_gaps(self):
        """Tests the case wherein there is multiple dates in the database, separated by a few months."""
        # Set up sample count mocking.
        fetch_all_mock = mock_fetch_all_sample_count(
            [
                {'project': 0, 'year': 2025, 'month': 2, 'count': 100},
                {'project': 0, 'year': 2025, 'month': 4, 'count': 150},
                {'project': 0, 'year': 2025, 'month': 7, 'count': 170},
            ]
        )
        self.sample_table.connection.fetch_all = mock.AsyncMock(
            side_effect=fetch_all_mock
        )
        result = await self.sample_table.get_monthly_samples_count_per_project()

        month_costs = result[
            0
        ]  # Retrieve the cost/month for the given test project ID.
        self.assertDictEqual(
            month_costs,
            {
                date(year=2025, month=2, day=1): 100,
                date(year=2025, month=3, day=1): 100,
                date(year=2025, month=4, day=1): 250,
                date(year=2025, month=5, day=1): 250,
                date(year=2025, month=6, day=1): 250,
                date(year=2025, month=7, day=1): 420,
                date(year=2025, month=8, day=1): 420,
                date(year=2025, month=9, day=1): 420,
            },
        )

    @run_as_sync
    @mock_today(date(year=2025, month=9, day=1))
    async def test_only_this_month(self):
        """Tests the case wherein the only record in the database is the current month."""
        # Set up sample count mocking.
        fetch_all_mock = mock_fetch_all_sample_count(
            [
                {'project': 0, 'year': 2025, 'month': 9, 'count': 170},
            ]
        )
        self.sample_table.connection.fetch_all = mock.AsyncMock(
            side_effect=fetch_all_mock
        )
        result = await self.sample_table.get_monthly_samples_count_per_project()

        month_costs = result[
            0
        ]  # Retrieve the cost/month for the given test project ID.
        self.assertDictEqual(month_costs, {date(year=2025, month=9, day=1): 170})



import unittest
from test.data.generate_data import QUERY_ENUMS, QUERY_SG_ID

from api.graphql.schema import schema  # type: ignore
from metamist.graphql import configure_sync_client, validate


class ValidateGenerateDataQueries(unittest.TestCase):
    """Validate generate data queries"""

    def test_validate_queries(self):
        """Validate queries"""
        client = configure_sync_client(schema=schema.as_str(), auth_token='FAKE')
        validate(QUERY_ENUMS, client=client)
        validate(QUERY_SG_ID, client=client)



"""CLI entry point for reviewing audit results."""

import click

from metamist.audit.data_access.gcs_data_access import GCSDataAccess
from metamist.audit.models import AuditReportEntry, ReviewResult
from metamist.audit.services import BucketAuditLogger, Reporter

from cpg_utils import to_path

ACTIONS = ['DELETE', 'INGEST', 'REVIEW']


def review_audit_report(
    dataset: str,
    results_folder: str,
    report: str = 'files_to_review',
    action: str = 'DELETE',
    comment: str = '',
    filter_expressions: list[str] | None = None,
):
    """Review files listed in the audit results."""
    audit_logs = BucketAuditLogger(dataset, 'audit_review')

    gcs_data = GCSDataAccess(dataset)
    reporter = Reporter(gcs_data, audit_logs, results_folder)

    audit_logs.info_nl(f"Reading report '{report}'".center(50, '~'))
    rows = reporter.get_report_rows_from_name(report)

    rows = reporter.filter_rows(filter_expressions, rows)
    review_result = review_rows(rows, action, comment, audit_logs)
    audit_logs.info_nl(
        f'Added comments to {len(review_result.reviewed_files)} / {len(rows)} files.'
    )
    if review_result.reviewed_files:
        reporter.write_reviewed_files_report(review_result)

    audit_logs.info_nl('Review complete!')


def review_rows(
    rows: list[AuditReportEntry],
    action: str,
    comment: str,
    audit_logs: BucketAuditLogger,
) -> ReviewResult:
    """Review files from audit results based on filters and comments."""
    action = action.upper()
    if action not in ACTIONS:
        raise ValueError(f'Invalid action: {action}. Must be one of {ACTIONS}.')
    audit_logs.info_nl(f"Annotating rows with comment: '{comment}'")
    reviewed_rows: list[AuditReportEntry] = []
    for row in rows:
        file_path = to_path(row.filepath)
        if not file_path.exists():
            audit_logs.warning(f'File {file_path} does not exist, skipping review.')
            continue

        row.update_action(action)
        row.update_review_comment(comment)
        reviewed_rows.append(row)

    return ReviewResult(reviewed_files=reviewed_rows)


@click.command()
@click.option('--dataset', '-d', required=True, help='Dataset name for logging context')
@click.option(
    '--results-folder', '-r', help='Path to the folder containing audit results files'
)
@click.option(
    '--action', '-a', default='DELETE', help='Action to perform on reviewed files.'
)
@click.option(
    '--comment', '-c', required=True, help='Comment to add to reviewed files.'
)
@click.option(
    '--filter',
    '-f',
    'filter_expressions',
    multiple=True,
    help='Filter expression like "SG Type==exome" or "File Path contains 2025-06-01"',
)
def main(
    dataset: str,
    results_folder: str,
    action: str,
    comment: str,
    filter_expressions: tuple = (),
):
    """Main function to review objects."""
    review_audit_report(
        dataset=dataset,
        results_folder=results_folder,
        action=action,
        comment=comment,
        filter_expressions=list(filter_expressions),
    )


if __name__ == '__main__':
    main()  # pylint: disable=no-value-for-parameter



import dataclasses
import unittest
from datetime import datetime
from enum import Enum
from typing import Any

from google.cloud import bigquery

from db.python.tables.bq.generic_bq_filter import GenericBQFilter
from db.python.tables.bq.generic_bq_filter_model import GenericBQFilterModel


@dataclasses.dataclass(kw_only=True)
class GenericBQFilterTest(GenericBQFilterModel):
    """Test model for GenericBQFilter"""

    test_string: GenericBQFilter[str] | None = None
    test_int: GenericBQFilter[int] | None = None
    test_float: GenericBQFilter[float] | None = None
    test_dt: GenericBQFilter[datetime] | None = None
    test_dict: dict[str, GenericBQFilter[str]] | None = None
    test_enum: GenericBQFilter[Enum] | None = None
    test_any: Any | None = None


class BGFilterTestEnum(str, Enum):
    """Simple Enum classs"""

    ID = 'id'
    VALUE = 'value'


class TestGenericBQFilter(unittest.TestCase):
    """Test generic filter SQL generation"""

    def test_basic_no_override(self):
        """Test that the basic filter converts to SQL as expected"""
        filter_ = GenericBQFilterTest(test_string=GenericBQFilter(eq='test'))
        sql, values = filter_.to_sql()

        self.assertEqual('test_string = @test_string_eq', sql)
        self.assertDictEqual(
            {
                'test_string_eq': bigquery.ScalarQueryParameter(
                    'test_string_eq', 'STRING', 'test'
                )
            },
            values,
        )

    def test_basic_override(self):
        """Test that the basic filter with an override converts to SQL as expected"""
        filter_ = GenericBQFilterTest(test_string=GenericBQFilter(eq='test'))
        sql, values = filter_.to_sql({'test_string': 't.string'})

        self.assertEqual('t.string = @t_string_eq', sql)
        self.assertDictEqual(
            {
                't_string_eq': bigquery.ScalarQueryParameter(
                    't_string_eq', 'STRING', 'test'
                )
            },
            values,
        )

    def test_single_string(self):
        """
        Test that a single value filtered using the "in" operator
        gets converted to an eq operation
        """
        filter_ = GenericBQFilterTest(test_string=GenericBQFilter(in_=['test']))
        sql, values = filter_.to_sql()

        self.assertEqual('test_string = @test_string_in_eq', sql)
        self.assertDictEqual(
            {
                'test_string_in_eq': bigquery.ScalarQueryParameter(
                    'test_string_in_eq', 'STRING', 'test'
                )
            },
            values,
        )

    def test_single_int(self):
        """
        Test that values filtered using the "in" operator convert as expected
        """
        value = 123
        filter_ = GenericBQFilterTest(test_int=GenericBQFilter(gt=value))
        sql, values = filter_.to_sql()

        self.assertEqual('test_int > @test_int_gt', sql)
        self.assertDictEqual(
            {
                'test_int_gt': bigquery.ScalarQueryParameter(
                    'test_int_gt', 'INT64', value
                )
            },
            values,
        )

    def test_single_float(self):
        """
        Test that values filtered using the "in" operator convert as expected
        """
        value = 123.456
        filter_ = GenericBQFilterTest(test_float=GenericBQFilter(gte=value))
        sql, values = filter_.to_sql()

        self.assertEqual('test_float >= @test_float_gte', sql)
        self.assertDictEqual(
            {
                'test_float_gte': bigquery.ScalarQueryParameter(
                    'test_float_gte', 'FLOAT64', value
                )
            },
            values,
        )

    def test_single_datetime(self):
        """
        Test that values filtered using the "in" operator convert as expected
        """
        datetime_str = '2021-10-08 01:02:03'
        value = datetime.strptime(datetime_str, '%Y-%m-%d %H:%M:%S')
        filter_ = GenericBQFilterTest(test_dt=GenericBQFilter(lt=value))
        sql, values = filter_.to_sql()

        self.assertEqual('test_dt < TIMESTAMP(@test_dt_lt)', sql)
        self.assertDictEqual(
            {
                'test_dt_lt': bigquery.ScalarQueryParameter(
                    'test_dt_lt', 'STRING', datetime_str
                )
            },
            values,
        )

    def test_single_enum(self):
        """
        Test that values filtered using the "in" operator convert as expected
        """
        value = BGFilterTestEnum.ID
        filter_ = GenericBQFilterTest(test_enum=GenericBQFilter(lte=value))
        sql, values = filter_.to_sql()

        self.assertEqual('test_enum <= @test_enum_lte', sql)
        self.assertDictEqual(
            {
                'test_enum_lte': bigquery.ScalarQueryParameter(
                    'test_enum_lte', 'STRING', value.value
                )
            },
            values,
        )

    def test_in_multiple_int(self):
        """
        Test that values filtered using the "in" operator convert as expected
        """
        value = [1, 2]
        filter_ = GenericBQFilterTest(test_int=GenericBQFilter(in_=value))
        sql, values = filter_.to_sql()

        self.assertEqual('test_int IN UNNEST(@test_int_in)', sql)
        self.assertDictEqual(
            {
                'test_int_in': bigquery.ArrayQueryParameter(
                    'test_int_in', 'INT64', value
                )
            },
            values,
        )

    def test_in_multiple_float(self):
        """
        Test that values filtered using the "in" operator convert as expected
        """
        value = [1.0, 2.0]
        filter_ = GenericBQFilterTest(test_float=GenericBQFilter(in_=value))
        sql, values = filter_.to_sql()

        self.assertEqual('test_float IN UNNEST(@test_float_in)', sql)
        self.assertDictEqual(
            {
                'test_float_in': bigquery.ArrayQueryParameter(
                    'test_float_in', 'FLOAT64', value
                )
            },
            values,
        )

    def test_in_multiple_str(self):
        """
        Test that values filtered using the "in" operator convert as expected
        """
        value = ['A', 'B']
        filter_ = GenericBQFilterTest(test_string=GenericBQFilter(in_=value))
        sql, values = filter_.to_sql()

        self.assertEqual('test_string IN UNNEST(@test_string_in)', sql)
        self.assertDictEqual(
            {
                'test_string_in': bigquery.ArrayQueryParameter(
                    'test_string_in', 'STRING', value
                )
            },
            values,
        )

    def test_nin_multiple_str(self):
        """
        Test that values filtered using the "in" operator convert as expected
        """
        value = ['A', 'B']
        filter_ = GenericBQFilterTest(test_string=GenericBQFilter(nin=value))
        sql, values = filter_.to_sql()

        self.assertEqual('test_string NOT IN UNNEST(@test_string_nin)', sql)
        self.assertDictEqual(
            {
                'test_string_nin': bigquery.ArrayQueryParameter(
                    'test_string_nin', 'STRING', value
                )
            },
            values,
        )

    def test_in_and_eq_multiple_str(self):
        """
        Test that values filtered using the "in" operator convert as expected
        """
        value = ['A']
        filter_ = GenericBQFilterTest(test_string=GenericBQFilter(in_=value, eq='B'))
        sql, values = filter_.to_sql()

        self.assertEqual(
            'test_string = @test_string_eq AND test_string = @test_string_in_eq',
            sql,
        )
        self.assertDictEqual(
            {
                'test_string_eq': bigquery.ScalarQueryParameter(
                    'test_string_eq', 'STRING', 'B'
                ),
                'test_string_in_eq': bigquery.ScalarQueryParameter(
                    'test_string_in_eq', 'STRING', 'A'
                ),
            },
            values,
        )

    def test_fail_none_in_tuple(self):
        """
        Test that values filtered using the "in" operator convert as expected
        """
        value = (None,)

        # check if ValueError is raised
        with self.assertRaises(ValueError) as context:
            filter_ = GenericBQFilterTest(test_any=value)
            filter_.to_sql()

        self.assertTrue(
            'There is very likely a trailing comma on the end of '
            'GenericBQFilterTest.test_any. '
            'If you actually want a tuple of length one with the value = (None,), '
            'then use dataclasses.field(default_factory=lambda: (None,))'
            in str(context.exception)
        )



from unittest import TestCase

from models.models import (
    PRIMARY_EXTERNAL_ORG,
    ParticipantUpsert,
    ParticipantUpsertInternal,
    SampleUpsert,
    SampleUpsertInternal,
)
from models.utils.sample_id_format import sample_id_format


class TestParticipantModels(TestCase):
    """Test participant model conversions"""

    def test_participant_to_internal_basic(self):
        """Test converting a basic participant to internal model"""
        external = ParticipantUpsert(external_ids={PRIMARY_EXTERNAL_ORG: 'hey-hey'})
        internal = external.to_internal()

        self.assertIsInstance(internal, ParticipantUpsertInternal)
        self.assertDictEqual({PRIMARY_EXTERNAL_ORG: 'hey-hey'}, internal.external_ids)

    def test_participant_to_external_basic(self):
        """Test converting a basic participant to external model"""
        internal = ParticipantUpsertInternal(
            id=1, external_ids={PRIMARY_EXTERNAL_ORG: 'hey-hey'}
        )
        external = internal.to_external()

        self.assertIsInstance(external, ParticipantUpsert)
        self.assertEqual(1, external.id)
        self.assertDictEqual({PRIMARY_EXTERNAL_ORG: 'hey-hey'}, external.external_ids)


class TestSampleModels(TestCase):
    """Test sample model conversions"""

    def test_sample_to_internal_basic(self):
        """Test converting a basic sample to internal model"""
        external = SampleUpsert(external_ids={PRIMARY_EXTERNAL_ORG: 'hey-hey'})
        internal = external.to_internal()

        self.assertIsInstance(internal, SampleUpsertInternal)
        self.assertDictEqual({PRIMARY_EXTERNAL_ORG: 'hey-hey'}, internal.external_ids)

    def test_sample_to_external_basic(self):
        """Test converting a basic sample to external model"""
        internal = SampleUpsertInternal(
            id=1, external_ids={PRIMARY_EXTERNAL_ORG: 'hey-hey'}
        )
        external = internal.to_external()

        self.assertIsInstance(external, SampleUpsert)
        self.assertEqual(sample_id_format(1), external.id)
        self.assertDictEqual({PRIMARY_EXTERNAL_ORG: 'hey-hey'}, external.external_ids)
